/***************************************************************
 * The mergesort implementations of this file are copied from
 * S. Mortensen, Refining the pure-C Cost Model, 
 * M. Sc. Thesis, Department of Computer Science, 
 * University of Copenhagen (2001). 
 ***************************************************************/

#ifndef KT_SORTING
#define KT_SORTING
#include <algorithm>
#include <iterator>

unsigned int* tmp = new unsigned int[MAXSIZE];
const size_t threshold = 3;

namespace kt { 
  namespace two_way { 
    template<class RandomIterator, 
	     class OutputIterator> 
    OutputIterator merge(RandomIterator b1, 
			 RandomIterator e1,
			 RandomIterator b2,
			 RandomIterator e2,
			 OutputIterator dest) {
      if (e2[-1] < e1[-1]) { 
	std::swap(b1, b2); 
	std::swap(e1, e2);
      } 
      typename std::iterator_traits<RandomIterator>::value_type v1, v2, tv1, tv2; 
      RandomIterator t1, t2; 
      v1 = *b1; 
      v2 = *b2; 
      t1 = e1; 
      t1 = t1 - 1;
      tv1 = *t1;
      t2 = e2;
      t2 = t2 - 1;
      tv2 = *t2; 
      if (tv1 > tv2) goto test_1;
      goto test_2; 
    out_1: b1 = b1 + 1;
      *dest = v1; 
      dest = dest + 1; 
      v1 = *b1; 
    test_1: if (v1 <= v2) goto out_1;
      b2 = b2 + 1;
      *dest = v2; 
      dest = dest + 1; 
      v2 = *b2; 
      if (b2 < e2) goto test_1; /* hint: taken */ 
    copy_1: b1 = b1 + 1; 
      *dest = v1; 
      dest = dest + 1; 
      v1 = *b1;
      if (b1 < e1) goto copy_1; /* hint: taken */ 
      goto end; 
    out_2: b2 = b2 + 1; 
      *dest = v2;
      dest = dest + 1;
      v2 = *b2; 
    test_2: if (v1 > v2) goto out_2; 
      b1 = b1 + 1; 
      *dest = v1; 
      dest = dest + 1;
      v1 = *b1; 
      if (b1 < e1) goto test_2; /* hint: taken */ 
    copy_2: b2 = b2 + 1;
      *dest = v2; 
      dest = dest + 1; 
      v2 = *b2; 
      if (b2 < e2) goto copy_2; /* hint: taken */ 
    end: return dest;
    } 

    template<class RandomIterator> 
    void sort(RandomIterator begin,
	      RandomIterator end,
	      RandomIterator dest) {
      typename std::iterator_traits<RandomIterator>::difference_type n = end - begin; 
      if (n > threshold) { 
	RandomIterator mid = begin + n/2; 
	RandomIterator dest_mid = dest + n/2;
	sort_to(begin, mid, dest); 
	sort_to(mid, end, dest_mid);
	merge(dest, dest_mid, dest_mid, dest+n, begin);
      } else { 
	std::__insertion_sort(begin, end, std::less<typename std::iterator_traits<RandomIterator>::value_type>()); 
      } 
    }
    
    template<class RandomIterator> 
    void sort_to(RandomIterator begin, RandomIterator end, RandomIterator dest) { 
      typename std::iterator_traits<RandomIterator>::difference_type n = end - begin;
      if (n > threshold) { 
	RandomIterator mid = begin + n/2;
	RandomIterator dest_mid = dest + n/2;
	sort(begin, mid, dest); 
	sort(mid, end, dest);
	merge(begin, mid, mid, end, dest); 
      } else { 
	std::__insertion_sort( begin, end, std::less<typename std::iterator_traits<RandomIterator>::value_type>()); 
	memcpy(dest, begin, (end - begin)*sizeof(typename std::iterator_traits<RandomIterator>::value_type)); 
      }
    } 

    void sort(unsigned int* begin, unsigned int* end) { 
      sort(begin, end, tmp);
    } 
  } 
  
  namespace three_way { 
    template<class RandomIterator, class OutputIterator> 
    OutputIterator merge(RandomIterator b1,
			 RandomIterator e1, 
			 RandomIterator b2, 
			 RandomIterator e2,
			 RandomIterator b3,
			 RandomIterator e3,
			 OutputIterator dest) {
      RandomIterator dest0 = dest; 
      if (e2[-1] < e1[-1]) { 
	std::swap(b1, b2); 
	std::swap(e1, e2);
      } 
      if (e3[-1] < e1[-1]) { 
	std::swap(b1, b3);
	std::swap(e1, e3); 
      } 
      typename std::iterator_traits<RandomIterator>::value_type v1, v2, v3; v1 = *b1;
      v2 = *b2;
      v3 = *b3;
      if (v2 <= v3) goto test_21;
      goto test_31;
    out_2: b2 = b2 + 1;
      *dest = v2;
      dest = dest + 1;
      v2 = *b2;
      if (v2 > v3) goto test_31;
    test_21: /* invariant: v2 <= v3 */ 
      if (v2 < v1) goto out_2; 
      b1 = b1 + 1; 
      *dest = v1;
      dest = dest + 1;
      v1 = *b1; 
      if (b1 < e1) goto test_21; /* hint: branch taken */ 
      goto exit; 
    out_3: b3 = b3 + 1; 
      *dest = v3; 
      dest = dest + 1; 
      v3 = *b3; 
      if (v2 <= v3) goto test_21; 
    test_31: /* invariant: v2 > v3 */ 
      if (v3 < v1) goto out_3;
      b1 = b1 + 1; 
      *dest = v1; 
      dest = dest + 1; 
      v1 = *b1; 
      if (b1 < e1) goto test_31; /* hint: branch taken */ 
    exit: /* final 2-way merge step */ 
      //      assert_sorted(dest0, dest);
      return two_way::merge(b2, e2, b3, e3, dest);
    } 

    template<class RandomIterator> 
    void sort(RandomIterator begin, RandomIterator end, RandomIterator dest)
    {
      typename std::iterator_traits<RandomIterator>::difference_type n = end - begin; 
      if (n > threshold) { 
	typename std::iterator_traits<RandomIterator>::value_type third = n/3;
	RandomIterator mid1 = begin + third; 
	RandomIterator dest_mid1 = dest + third; 
	RandomIterator mid2 = begin + 2*third; 
	RandomIterator dest_mid2 = dest + 2*third; 
	sort_to(begin, mid1, dest); 
	sort_to(mid1, mid2, dest_mid1); 
	sort_to(mid2, end, dest_mid2);
	merge(dest, dest_mid1, dest_mid1, dest_mid2, dest_mid2, dest+n, begin); 
      } else {
	std::__insertion_sort( begin, end, std::less<typename std::iterator_traits<RandomIterator>::value_type>()); 
      }
    } 

    template<class RandomIterator> 
    void sort_to(RandomIterator begin, RandomIterator end, RandomIterator dest) {
      typename std::iterator_traits<RandomIterator>::difference_type n = end - begin; 
      if (n > threshold) { 
	typename std::iterator_traits<RandomIterator>::value_type third = n/3; 
	RandomIterator mid1 = begin + third; 
	RandomIterator dest_mid1 = dest + third; 
	RandomIterator mid2 = begin + 2*third; 
	RandomIterator dest_mid2 = dest + 2*third; 
	sort(begin, mid1, dest); 
	sort(mid1, mid2, dest);
	sort(mid2, end, dest); 
	merge(begin, mid1, mid1, mid2, mid2, end, dest);
      } else { 
	std::__insertion_sort(begin, end, std::less<typename std::iterator_traits<RandomIterator>::value_type>()); 
	memcpy( dest, begin, (end - begin)*sizeof(typename std::iterator_traits<RandomIterator>::value_type));
      } 
    }

    void sort(unsigned int* begin, unsigned int* end) { 
      sort(begin, end, tmp);
    } 
  } 

  namespace four_way { 
    template<class RandomIterator, 
	     class OutputIterator> 
    OutputIterator merge(RandomIterator b1, 
			 RandomIterator e1,
			 RandomIterator b2,
			 RandomIterator e2,
			 RandomIterator b3,
			 RandomIterator e3,
			 RandomIterator b4,
			 RandomIterator e4,
			 OutputIterator dest)
    {
      OutputIterator dest0 = dest; 
      //      assert_sorted(b1, e1); 
      //      assert_sorted(b2, e2);
      //      assert_sorted(b3, e3); 
      //      assert_sorted(b4, e4);
      if (e2[-1] < e1[-1]) { 
	std::swap(b1, b2); 
	std::swap(e1, e2);
      } 
      if (e3[-1] < e1[-1]) {
	std::swap(b1, b3);
	std::swap(e1, e3); 
      }
      if (e4[-1] < e1[-1]) {
	std::swap(b1, b4); 
	std::swap(e1, e4);
      } 
      typename std::iterator_traits<RandomIterator> ::value_type v1, v2, v3, v4;
      v1 = *b1;
      v2 = *b2;
      v3 = *b3;
      v4 = *b4;
      if (v2 <= v3 && v4 < v1) goto test_24;
      if (v2 <= v3 && v4 >= v1) goto test_21;
      if (v2 > v3 && v4 < v1) goto test_34;
      goto test_31;
    out_24: b2 = b2 + 1;
      *dest = v2; 
      dest = dest + 1; 
      v2 = *b2; 
      if (v2 > v3) goto test_34; 
    test_24: /* invariant: v2 <= v3 && v4 < v1 */ 
      //      assert( v2 <= v3 && v4 < v1 ); 
      if (v2 <= v4) goto out_24; 
      b4 = b4 + 1; 
      *dest = v4; 
      dest = dest + 1; 
      v4 = *b4; 
      if (v4 < v1) goto test_24; /* invariant: v2 <= v3 && v4 >= v1 */ 
      //      assert( v2 <= v3 && v4 >= v1 ); 
      if (v2 >= v1) goto out_12;
    out_21: b2 = b2 + 1;
      *dest = v2;
      dest = dest + 1;
      v2 = *b2; 
      if (v2 > v3) goto test_31;
    test_21: /* invariant: v2 <= v3 && v4 >= v1 */ 
      //      assert( v2 <= v3 && v4 >= v1 ); 
      if (v2 < v1) goto out_21;
    out_12: b1 = b1 + 1; 
      *dest = v1; 
      dest = dest + 1;
      v1 = *b1; 
      if (b1 >= e1) goto exit; /* hint: not taken */ 
      if (v4 >= v1) goto test_21;
      goto test_24; 
    out_34: b3 = b3 + 1;
      *dest = v3; 
      dest = dest + 1; 
      v3 = *b3;
      if (v2 <= v3) goto test_24;
    test_34: /* invariant: v2 > v3 && v4 < v1 */ 
      //      assert( v2 > v3 && v4 < v1 ); 
      if (v3 <= v4) goto out_34;
      b4 = b4 + 1; 
      *dest = v4; 
      dest = dest + 1;
      v4 = *b4;
      if (v4 < v1) goto test_34;/* invariant: v2 > v3 && v4 >= v1 */
      //      assert( v2 > v3 && v4 >= v1 );
      if (v3 >= v1) goto out_13;
    out_31: b3 = b3 + 1; 
      *dest = v3; 
      dest = dest + 1; 
      v3 = *b3; 
      if (v2 <= v3) goto test_21; 
    test_31: /* invariant: v2 > v3 && v4 >= v1 */ 
      //      assert( v2 > v3 && v4 >= v1 ); 
      if (v3 < v1) goto out_31; 
    out_13: b1 = b1 + 1; 
      *dest = v1; 
      dest = dest + 1; 
      v1 = *b1; 
      if (b1 >= e1) goto exit; /* hint: not taken */ 
      if (v4 >= v1) goto test_31; 
      goto test_34;
    exit: /* merge remaining three subarrays */ 
      //      assert_sorted(dest0, dest);
      return three_way::merge(b2, e2, b3, e3, b4, e4, dest); 
    }
    
    template<class RandomIterator>
    void sort(RandomIterator begin,
	      RandomIterator end, 
	      RandomIterator dest) { 
      typename std::iterator_traits<RandomIterator>::difference_type n = end - begin; 
      if (n > threshold) { 
	typename std::iterator_traits<RandomIterator>::value_type fourth = n/4; 
	RandomIterator mid1 = begin + 1*fourth; 
	RandomIterator dest_mid1 = dest + 1*fourth; 
	RandomIterator mid2 = begin + 2*fourth; 
	RandomIterator dest_mid2 = dest + 2*fourth; 
	RandomIterator mid3 = begin + 3*fourth; 
	RandomIterator dest_mid3 = dest + 3*fourth; 
	sort_to(begin, mid1, dest); 
	sort_to(mid1, mid2, dest_mid1); 
	sort_to(mid2, mid3, dest_mid2); 
	sort_to(mid3, end, dest_mid3); 
	merge(dest, dest_mid1, dest_mid1, dest_mid2, dest_mid2, dest_mid3, dest_mid3, dest+n, begin);
      } else {
	std::__insertion_sort( begin, end, std::less<typename std::iterator_traits<RandomIterator>::value_type>()); 
      }
      //      assert_sorted(begin, end);
    }
    
    template<class RandomIterator>
    void sort_to(RandomIterator begin, 
		 RandomIterator end, 
		 RandomIterator dest) { 
      typename std::iterator_traits<RandomIterator>::difference_type n = end - begin; 
      if (n > threshold) { 
	typename std::iterator_traits<RandomIterator>::value_type fourth = n/4; 
	RandomIterator mid1 = begin + 1*fourth; 
	RandomIterator dest_mid1 = dest + 1*fourth; 
	RandomIterator mid2 = begin + 2*fourth; 
	RandomIterator dest_mid2 = dest + 2*fourth; 
	RandomIterator mid3 = begin + 3*fourth; 
	RandomIterator dest_mid3 = dest + 3*fourth; 
	sort(begin, mid1, dest); 
	sort(mid1, mid2, dest);
	sort(mid2, mid3, dest);
	sort(mid3, end, dest);
	merge(begin, mid1, mid1, mid2, mid2, mid3, mid3, end, dest);
      } else {
	std::__insertion_sort( begin, end, std::less<typename std::iterator_traits<RandomIterator>::value_type>()); 
	memcpy( dest,begin, (end - begin)*sizeof(typename std::iterator_traits<RandomIterator>::value_type));
      } 
      //      assert_sorted(dest, dest+n);
    } 
    
    void sort(unsigned int* begin, unsigned int* end) { 
      sort(begin, end, tmp);
    } 
  }
}
#endif // KT_SORTING
